import numpy as np

from collections import namedtuple
from extra_data.exceptions import PropertyNameError
from .agipd_base import (
    AgipdLitFrameFinderBase, Agipd1g, Agipd2g, ReferenceDelay,
    EVENT_CODE, DESTINATION, AGIPD_MAX_CELL, SA1, SA2, SA3,
)


# Exceptions
class LitFrameFinderError(RuntimeError):
    pass


class SourceNotFound(LitFrameFinderError):
    pass


class BunchPatternNotFound(SourceNotFound):
    pass


class DetectorNotFound(SourceNotFound):
    pass


class TriggerNotFound(SourceNotFound):
    pass


class NoReferenceDelay(LitFrameFinderError):
    pass


AttributeValue = namedtuple('AttributeValue', ['value'])
AVal = AttributeValue


class Agipd1gOffline(Agipd1g):
    def read_run_values(self, dc):
        srcname = self.detector_control_id

        AgipdBunchStructure = namedtuple('Agipd1BunchStructure', [
            'nPulses', 'firstPulse', 'repetitionRate'
        ])
        AgipdDetectorCtrl = namedtuple('Agipd1DetectorCtrl', [
            'bunchStructure', 'patternType', 'patternTypeIndex',
            'integrationTime', 't0Delay'
        ])
        self.det = AgipdDetectorCtrl(
            bunchStructure=AgipdBunchStructure(
                nPulses=AVal(dc.get_run_value(
                    srcname, 'bunchStructure.nPulses')),
                firstPulse=AVal(dc.get_run_value(
                    srcname, 'bunchStructure.firstPulse')),
                repetitionRate=AVal(dc.get_run_value(
                    srcname, 'bunchStructure.repetitionRate')),
            ),
            patternType=AVal(dc.get_run_value(
                srcname, 'patternType')),
            patternTypeIndex=AVal(dc.get_run_value(
                srcname, 'patternTypeIndex')),
            integrationTime=AVal(dc.get_run_value(
                srcname, 'integrationTime')),
            t0Delay=AVal(dc.get_run_value(
                srcname, 't0Delay')),
        )
        return self.det


class Agipd2gOffline(Agipd2g):
    def read_run_values(self, dc):
        srcname = self.detector_control_id

        AgipdBunchStructure = namedtuple('Agipd2BunchStructure', [
            'nPulses', 'repetitionRate'
        ])
        AgipdDetectorCtrl = namedtuple('Agipd2DetectorCtrl', [
            'bunchStructure', 'expType', 'expTypeIndex',
            'integrationTime',
        ])
        self.det = AgipdDetectorCtrl(
            bunchStructure=AgipdBunchStructure(
                nPulses=AVal(dc.get_run_value(
                    srcname, 'bunchStructure.nPulses')),
                repetitionRate=AVal(dc.get_run_value(
                    srcname, 'bunchStructure.repetitionRate')),
            ),
            expType=AVal(dc.get_run_value(
                srcname, 'expType')),
            expTypeIndex=AVal(dc.get_run_value(
                srcname, 'expTypeIndex')),
            integrationTime=AVal(dc.get_run_value(
                srcname, 'integrationTime')),
        )
        return self.det


AGIPDGEN_OFFLINE = {
    "AgipdComposite": Agipd1gOffline,
    "Agipd2Composite": Agipd2gOffline,
}

Trigger = namedtuple('Trigger', ['deviceId', 'name'])
Motor = namedtuple('Motor', ['deviceId', 'positionProperty',
                             'unitConversionCoef'])


class AgipdLitFrameFinderOffline(AgipdLitFrameFinderBase):

    def __init__(self, dev_id, dc, ref_delays={}, **kwargs):
        self.dev_id = dev_id
        self.dc = dc
        self._reference = ref_delays

        for a, v in kwargs.items():
            setattr(self, a, AVal(v))

        if dev_id in dc.control_sources:
            try:
                self._configure(dc, dev_id)
            except PropertyNameError:
                self._configure_v03(dc, dev_id)

        inst = self.instrumentName.value
        self.dst = DESTINATION[inst]

        agipd_class = AGIPDGEN_OFFLINE.get(self._get_agipd_classid(dc))
        if agipd_class is None:
            raise ValueError("Unknown class of AgipdComposite device")
        self._agipd = agipd_class(self.detector_control_id, self)

        self.shutters_are_open = True

    def _configure_v03(self, dc, dev_id):
        def get_property(key):
            return AVal(dc.get_run_value(dev_id, key))

        def get_value(key):
            return dc.get_run_value(dev_id, key)

        self.detector_control_id = get_value('DetectorCtrl')
        self.alignMethod = get_property('alignMethod')
        self.instrumentName = get_property('instrumentName')
        self.referenceFrame = get_property('referenceFrame')

        self.trigger = Trigger(
            deviceId=get_value('Timer'),
            name=get_property('trigger'),
        )
        self.motor = Motor(
            deviceId='',
            positionProperty='',
            unitConversionCoef=1,
        )

    def _configure(self, dc, dev_id):
        def get_property(key):
            return AVal(dc.get_run_value(dev_id, key))

        def get_value(key):
            return dc.get_run_value(dev_id, key)

        self.detector_control_id = get_value('detectorId')
        self.alignMethod = get_property('alignMethod')
        self.instrumentName = get_property('instrumentName')
        self.referenceFrame = get_property('referenceFrame')

        self.trigger = Trigger(
            deviceId=get_value('trigger.deviceId'),
            name=get_property('trigger.name'),
        )
        self.motor = Motor(
            deviceId=get_value('motor.deviceId'),
            positionProperty=get_property('motor.positionProperty'),
            unitConversionCoef=get_property('motor.unitConversionCoef'),
        )
        for event_name, event_code in EVENT_CODE.items():
            ref = ReferenceDelay(**dict([
                (key, get_value(f"referenceDelay.{event_name}.{key}"))
                for key in ReferenceDelay._fields
            ]))
            if ref.delay != -1:
                self._reference[event_code] = ref

    def _get_agipd_classid(self, dc):
        srcname = self.detector_control_id
        if srcname not in dc.control_sources:
            raise(DetectorNotFound(f"Source '{srcname}' is not found."))

        return dc.get_run_value(srcname, 'classId')

    def _read_run_values(self, dc):
        self.DetectorCtrl = self._agipd.read_run_values(dc)

        if self.alignMethod.value == "by reference delay":
            self._read_trigger_and_motor_values(dc)

    def _read_trigger_and_motor_values(self, dc):
        srcname = self.trigger.deviceId
        if srcname not in dc.control_sources:
            raise(TriggerNotFound(f"Source {srcname} is not found."))

        key = self.trigger.name.value
        self.trigger_event = dc.get_run_value(srcname, key + '.event')
        self.trigger_delay = dc.get_run_value(srcname, key + '.delay')

        self.ref = self._reference.get(self.trigger_event)
        if self.ref is None:
            raise(NoReferenceDelay("Reference delay for marco event "
                                   f"{self.trigger_event} is not set."))

        srcname = self.motor.deviceId
        if srcname in dc.control_sources:
            key = self.motor.positionProperty.value
            self.motor_position = dc.get_run_value(srcname, key)
        else:
            self.motor_position = None

    def _read_pp_decoder(self, dc):
        pp_src = dc[self.pp_decoder_id]
        trainId = pp_src['sase1/nPulses'].train_id_coordinates()
        bunch_pattern = np.zeros([trainId.size, 2700], dtype=np.uint32)
        n1 = pp_src['sase1/nPulses'].ndarray()
        p1 = pp_src['sase1/pulseIds'].ndarray()
        n2 = pp_src['sase2/nPulses'].ndarray()
        p2 = pp_src['sase2/pulseIds'].ndarray()
        n3 = pp_src['sase3/nPulses'].ndarray()
        p3 = pp_src['sase3/pulseIds'].ndarray()
        for i in range(trainId.size):
            bunch_pattern[i, p1[i, :n1[i]]] |= SA1.code
            bunch_pattern[i, p2[i, :n2[i]]] |= SA2.code
            bunch_pattern[i, p3[i, :n3[i]]] |= SA3.code
        return trainId, bunch_pattern

    def process(self, force=False):
        if not force and hasattr(self, 'r'):
            return self.r

        dc = self.dc
        self._read_run_values(dc)

        # bunch pattern
        if self.bunch_pattern_id in dc.instrument_sources:
            bp_keydata = dc[self.bunch_pattern_id, 'data.bunchPatternTable']
            bunch_pattern = bp_keydata.ndarray()
            trainId = bp_keydata.train_id_coordinates()
            bpsrc = "timeserver"
        elif self.pp_decoder_id in dc.control_sources:
            trainId, bunch_pattern = self._read_pp_decoder(dc)
            bpsrc = "pp_decoder"
        elif self._agipd.exp_type.value == 'XRay':
            raise(BunchPatternNotFound(
                f"Neither source '{self.bunch_pattern_id}' nor source "
                f"'{self.pp_decoder_id}' are found."))
        else:
            # only for dark runs
            bpsrc = "none"
            trainId = np.array(dc.train_ids, np.uint64)
            bunch_pattern = None

        # XGM
        if self.xgm_id not in dc.instrument_sources:
            xgm_flag = np.zeros(trainId.size, bool)
        else:
            xgm_en_keydata = dc[self.xgm_id, 'data.intensityTD']
            xgm_sig_keydata = dc[self.xgm_id, 'data.intensitySigmaTD']

            xgm_trainId = xgm_en_keydata.train_id_coordinates()
            xgm_sel = np.in1d(xgm_trainId, trainId)
            xgm_trainId = xgm_trainId[xgm_sel]

            xgm_intensityTD = xgm_en_keydata.ndarray()[xgm_sel]
            xgm_intensitySigmaTD = xgm_sig_keydata.ndarray()[xgm_sel]

            xgm_flag = np.in1d(trainId, xgm_trainId)
            xgm_map = np.roll(np.cumsum(xgm_flag), 1)
            xgm_map[0] = 0

        Output = namedtuple('Output', [
            'nFrame', 'detectorPulseId', 'nPulsePerFrame', 'energyPerFrame',
            'energySigma', 'dataFramePattern', 'masterPulseId', 'xgmPulseId'
        ])
        Result = namedtuple('Results', [
            'nPulse', 'nFrame', 'litFrames', 'nLitFrame', 'dataFrames',
            'nDataFrame', 'output', 'meta'
        ])
        Meta = namedtuple('Meta', ['litFrmDev', 'trainId'])
        n = trainId.shape[0]
        r = Result(
            nPulse=AVal(np.zeros(n, np.uint16)),
            nFrame=AVal(np.zeros(n, np.uint16)),
            nLitFrame=AVal(np.zeros(n, np.uint16)),
            litFrames=AVal(np.zeros([n, AGIPD_MAX_CELL], np.uint16)),
            nDataFrame=AVal(np.zeros(n, np.uint16)),
            dataFrames=AVal(np.zeros([n, AGIPD_MAX_CELL], np.uint16)),
            output=Output(
                nFrame=np.zeros(n, np.uint16),
                nPulsePerFrame=np.zeros([n, AGIPD_MAX_CELL], np.uint16),
                energyPerFrame=np.zeros([n, AGIPD_MAX_CELL], np.float32),
                energySigma=np.zeros([n, AGIPD_MAX_CELL], np.float32),
                dataFramePattern=np.zeros([n, AGIPD_MAX_CELL], bool),
                detectorPulseId=np.zeros([n, AGIPD_MAX_CELL], np.uint16),
                masterPulseId=np.zeros([n, 2700], np.uint16),
                xgmPulseId=np.zeros([n, 2700], np.uint16),
            ),
            meta=Meta(litFrmDev=f"offline/{bpsrc}", trainId=trainId),
        )

        Data = namedtuple('Data', ['data'])
        BunchPatternTable = namedtuple(
            'BunchPatternTable', ['bunchPatternTable'])
        XGM = namedtuple('XGM', ['intensityTD', 'intensitySigmaTD'])

        for no, tid in enumerate(trainId):
            if bunch_pattern is not None:
                bp = Data(Data(BunchPatternTable(AVal(bunch_pattern[no]))))

                if xgm_flag[no]:
                    xgm_no = xgm_map[no]
                    xgm = Data(Data(XGM(
                        intensityTD=AVal(xgm_intensityTD[xgm_no]),
                        intensitySigmaTD=AVal(xgm_intensitySigmaTD[xgm_no]),
                    )))
                else:
                    xgm = None

                (npulse, ncell,
                 agipd_pulse_ids, frames) = self.process_train(bp, xgm)
            else:
                (npulse, ncell,
                 agipd_pulse_ids, frames) = self.make_dark_run()

            # send pipeline (fast) data
            r.output.nFrame[no] = ncell
            r.output.detectorPulseId[no, :ncell] = agipd_pulse_ids
            r.output.nPulsePerFrame[no, :ncell] = frames.npulse
            r.output.energyPerFrame[no, :ncell] = frames.energy
            r.output.energySigma[no, :ncell] = frames.energy_std
            r.output.dataFramePattern[no, :ncell] = frames.hasdata
            nmaster = frames.master_ids.size
            r.output.masterPulseId[no, :nmaster] = frames.master_ids
            nxgm = frames.xgm_ids.size
            r.output.xgmPulseId[no, :nxgm] = frames.xgm_ids

            litframes = np.flatnonzero(frames.npulse)
            dataframes = np.flatnonzero(frames.hasdata)

            # update attributes (slow data)
            r.nPulse.value[no] = npulse
            r.nFrame.value[no] = ncell
            r.litFrames.value[no, :frames.nlit] = litframes
            r.nLitFrame.value[no] = frames.nlit
            r.dataFrames.value[no, :frames.ndata] = dataframes
            r.nDataFrame.value[no] = frames.ndata

        self.r = r
        return r

    def read(self):
        dc = self.dc
        inst_src = dc[self.dev_id + ':output']
        trainId = inst_src['data.nFrame'].train_id_coordinates()

        Output = namedtuple('Output', [
            'nFrame', 'detectorPulseId', 'nPulsePerFrame', 'energyPerFrame',
            'energySigma', 'dataFramePattern', 'masterPulseId', 'xgmPulseId'
        ])
        Result = namedtuple('Results', [
            'nPulse', 'nFrame', 'litFrames', 'nLitFrame', 'dataFrames',
            'nDataFrame', 'output', 'meta'
        ])
        Meta = namedtuple('Meta', ['litFrmDev', 'trainId'])
        out = Output(
            nFrame=inst_src['data.nFrame'].ndarray(),
            nPulsePerFrame=inst_src['data.nPulsePerFrame'].ndarray(),
            energyPerFrame=inst_src['data.energyPerFrame'].ndarray(),
            energySigma=inst_src['data.energySigma'].ndarray(),
            dataFramePattern=inst_src['data.dataFramePattern'].ndarray(),
            detectorPulseId=inst_src['data.detectorPulseId'].ndarray(),
            masterPulseId=inst_src['data.masterPulseId'].ndarray(),
            xgmPulseId=inst_src['data.xgmPulseId'].ndarray(),
        )

        n = trainId.size
        litframes = np.zeros([n, 352], dtype=np.uint16)
        dataframes = np.zeros([n, 352], dtype=np.uint16)
        for no, tid in enumerate(trainId):
            a = np.flatnonzero(out.nPulsePerFrame[no])
            litframes[no, :a.size] = a
            a = np.flatnonzero(out.dataFramePattern[no])
            dataframes[no, :a.size] = a

        r = Result(
            nPulse=AVal(out.nPulsePerFrame.sum(1)),
            nFrame=AVal(out.nFrame),
            nLitFrame=AVal((out.nPulsePerFrame != 0).sum(1)),
            litFrames=AVal(litframes),
            nDataFrame=AVal(out.dataFramePattern.sum(1)),
            dataFrames=AVal(dataframes),
            output=out,
            meta=Meta(litFrmDev=self.dev_id, trainId=trainId),
        )

        self.r = r
        return r

    def read_or_process(self):
        channel = self.dev_id + ':output'
        if channel in self.dc.instrument_sources:
            r = self.read()
            if len(r.meta.trainId):
                return r

        return self.process(force=True)


class AgipdLitFrameFinderConfigureMID:
    dev_id = "MID_EXP_AGIPD1M1/REDU/LITFRM"

    detector_control_id = "MID_EXP_AGIPD1M1/MDL/FPGA_COMP"
    bunch_pattern_id = "MID_RR_UTC/TSYS/TIMESERVER:outputBunchPattern"
    xgm_id = "SA2_XTD1_XGM/XGM/DOOCS:output"
    pp_decoder_id = "MID_RR_SYS/MDL/PULSE_PATTERN_DECODER"

    trigger = Trigger(
        deviceId="MID_EXP_SYS/TSYS/UTC-2-S4",
        name=AVal("backTrg3")
    )
    motor = Motor(
        deviceId="MID_AGIPD_MOTION/MDL/DOWNSAMPLER",
        positionProperty=AVal("t4EncoderPosition"),
        unitConversionCoef=AVal(1e-3),
    )
    alignMethod = AVal("by first pulse")
    referenceFrame = AVal(1)

    missedIntensityTD = AVal(10000)
    missedIntensitySigmaTD = AVal(0)

    instrumentName = AVal("MID")


class AgipdLitFrameFinderMID(AgipdLitFrameFinderConfigureMID,
                             AgipdLitFrameFinderOffline):
    def __init__(self, dc, dev_id=AgipdLitFrameFinderConfigureMID.dev_id,
                 ref_delays={}, **kwargs):
        super().__init__(self.dev_id, dc, ref_delays, **kwargs)


class AgipdLitFrameFinderConfigureSPB:
    dev_id = 'SPB_IRU_AGIPD1M1/REDU/LITFRM'

    detector_control_id = 'SPB_IRU_AGIPD1M1/MDL/FPGA_COMP'
    bunch_pattern_id = 'SPB_RR_SYS/TSYS/X2TIMER2:outputBunchPattern'
    xgm_id = 'SPB_XTD9_XGM/XGM/DOOCS:output'
    pp_decoder_id = 'SPB_RR_SYS/MDL/BUNCH_PATTERN'

    trigger = Trigger(
        deviceId="SPB_EXP_SYS/TSYS/UTC_1_S3",
        name=AVal("backTrg3")
    )
    motor = Motor(
        deviceId="",
        positionProperty=AVal(""),
        unitConversionCoef=AVal(1),
    )
    alignMethod = AVal('by first pulse')
    referenceFrame = AVal(1)

    missedIntensityTD = AVal(10000)
    missedIntensitySigmaTD = AVal(0)

    instrumentName = AVal('SPB')


class AgipdLitFrameFinderSPB(AgipdLitFrameFinderConfigureSPB,
                             AgipdLitFrameFinderOffline):
    def __init__(self, dc, dev_id=AgipdLitFrameFinderConfigureSPB.dev_id,
                 ref_delays={}, **kwargs):
        super().__init__(self.dev_id, dc, ref_delays, **kwargs)


class AgipdLitFrameFinderConfigureHED:
    dev_id = 'HED_EXP_AGIPD500K2G/REDU/LITFRM'

    detector_control_id = 'HED_EXP_AGIPD500K2G/MDL/FPGA_COMP'
    bunch_pattern_id = 'HED_RR_UTC/TSYS/TIMESERVER:outputBunchPattern'
    xgm_id = 'HED_XTD6_XGM/XGM/DOOCS:output'
    pp_decoder_id = 'SA2_BR_UTC/MDL/BUNCHPATTERN_DECODER'

    trigger = Trigger(
        deviceId="HED_EXP_SYS/TSYS/UTC-1-S2",
        name=AVal("backTrg3")
    )
    motor = Motor(
        deviceId="",
        positionProperty=AVal(""),
        unitConversionCoef=AVal(1),
    )
    alignMethod = AVal('by first pulse')
    referenceFrame = AVal(1)

    missedIntensityTD = AVal(10000)
    missedIntensitySigmaTD = AVal(0)

    instrumentName = AVal('HED')


class AgipdLitFrameFinderHED(AgipdLitFrameFinderConfigureHED,
                             AgipdLitFrameFinderOffline):
    def __init__(self, dc, dev_id=AgipdLitFrameFinderConfigureHED.dev_id,
                 ref_delays={}, **kwargs):
        super().__init__(self.dev_id, dc, ref_delays, **kwargs)


AgipdLitFrameFinderProfiles = {
    "MID": AgipdLitFrameFinderMID,
    "SPB": AgipdLitFrameFinderSPB,
    "HED": AgipdLitFrameFinderHED,
}


def find_litframe_finder_sources(dc):
    litfrm_finder_ids = [
        src for src, cls in (
            (src, dc.get_run_value(src, 'classId'))
            for src in dc.control_sources
        ) if cls == 'AgipdLitFrameFinder'
    ]
    return litfrm_finder_ids


def make_litframe_finder(inst, dc, dev_id=None, ref_delays={}):
    if inst not in AgipdLitFrameFinderProfiles:
        raise IndexError("Unexpected name of AgipdLitFrameFinder profile")

    litfrm_finder_cls = AgipdLitFrameFinderProfiles[inst]
    dev_ids = find_litframe_finder_sources(dc)

    if not dev_id:
        dev_id = litfrm_finder_cls.dev_id

    if dev_ids and dev_id not in dev_ids:
        dev_id = dev_ids[0]

    return litfrm_finder_cls(dc, dev_id, ref_delays=ref_delays)



import numpy as np


def steady_intervals(a):
    """Find intervals of consecutive identical values"""
    if len(a) > 0:
        i = np.flatnonzero(np.diff(a))
        i0 = 0
        intv = True
        for iN in i:
            if intv:
                yield (i0, iN)
                intv = False
            else:
                i0 = iN + 1
                intv = True

        if intv:
            yield (i0, len(a) - 1)


def find_intervals(ptrn, intrv):
    ix = np.flatnonzero(ptrn)
    if ix.size == 1:
        intrv.append((ix[0],))
    else:
        steps = np.diff(ix)
        for f0, fN in steady_intervals(steps):
            i0 = ix[f0]
            if f0 == fN:
                intrv.append((i0,))
            else:
                iN = ix[fN + 1] + 1
                s = steps[f0]
                intrv.append((i0, iN) if s == 1 else (i0, iN, s))
    return intrv


def litfrm_run_report(r):
    """Compute single run litframe report"""
    rep = []
    ptrnno = 0
    nfrm_uniq, nfrm_ix = np.unique(r.output.nFrame, return_inverse=True)
    for i in range(nfrm_uniq.size):
        nfrm = nfrm_uniq[i]
        npls_uniq, npls_ix = np.unique(
            r.output.nPulsePerFrame[nfrm_ix == i, :nfrm],
            return_inverse=True, axis=0
        )
        datafrm = r.output.dataFramePattern[nfrm_ix == i, :nfrm]
        tid = r.meta.trainId[nfrm_ix == i]
        one = tid.dtype.type(1)

        trsint = []
        for j in range(npls_uniq.shape[0]):
            trsint = find_intervals(npls_ix == j, trsint)

        for trs in sorted(trsint):
            t0 = trs[0]
            tN = trs[1] - 1 if len(trs) > 1 else t0
            trains = ((tid[t0], tid[tN] + one, trs[2]) if len(trs) == 3
                      else (tid[t0], tid[tN] + one))

            ptrn = datafrm[t0, :]
            ndata = ptrn.sum()
            npls = npls_uniq[npls_ix[t0], :].sum()

            frmint = find_intervals(ptrn, [])
            rep.append(dict(
                source=r.meta.litFrmDev, pattern_no=ptrnno, trains=trains,
                npulse=npls, ndataframe=ndata, nframe=nfrm, frames=frmint,
            ))
            ptrnno += 1

    return rep



import os
import csv
import argparse

from extra_data import RunDirectory
from extra_data.read_machinery import find_proposal
from extra_redu import (
    make_litframe_finder, BunchPatternNotFound, DetectorNotFound
)
from extra_redu.litfrm.utils import litfrm_run_report


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Generate litframe report for a proposal.')
    parser.add_argument('proposal', metavar='propno', type=int, nargs=1,
                        help='proposal Id')
    parser.add_argument('-o', '--csv', metavar='csv', type=str, nargs=1,
                        help='the name of output CSV-file')

    args = parser.parse_args()

    propno = args.proposal[0]
    csvname = f"litfrm_p{propno:06}.csv" if args.csv is None else args.csv[0]

    propdir = find_proposal(f"p{propno:06d}")
    inst = propdir.split('/')[4]
    print(f"Proposal No: {propno}")
    print(f"Proposal directory: {propdir}")
    print(f"Instrument: {inst}")

    runs = sorted([
        int(fn[1:]) for fn in os.listdir(os.path.join(propdir, "raw"))
        if fn[0] == 'r' and fn[1:].isdigit()
    ])

    f = open(csvname, 'w', newline='')
    fieldnames = ['runno', 'source', 'runtype', 'pattern_no', 'trains',
                  'npulse', 'ndataframe', 'nframe', 'frames', 'exception']
    writer = csv.DictWriter(
        f, fieldnames=fieldnames, delimiter=';', quotechar='"',
        quoting=csv.QUOTE_NONNUMERIC
    )
    writer.writeheader()

    for runno in runs:
        rundir = os.path.join(propdir, "raw", f"r{runno:04}")
        dc = RunDirectory(rundir)

        try:
            litfrm = make_litframe_finder(inst, dc)
            r = litfrm.process()
            runtype = litfrm.DetectorCtrl.patternType.value
            rep = litfrm_run_report(r)
            for r in rep:
                frmintf = ', '.join(
                    [':'.join([str(n) for n in slc]) for slc in r['frames']]
                )
                trsintf = ':'.join([str(n) for n in r['trains']])
                recfrm = r.copy()
                recfrm['frames'] = frmintf
                recfrm['runno'] = runno
                recfrm['runtype'] = runtype
                writer.writerow(recfrm)
                if r['pattern_no'] == 0:
                    print("{runno:4d} {source:30s} {runtype:6s}".format(
                        **recfrm), end='')
                else:
                    print(" "*42, end='')
                print(
                    ("{pattern_no:2d} {trsintf:25s} {npulse:4d} "
                     "{ndataframe:3d} {nframe:3d} [{frames}]"
                     ).format(trsintf=trsintf, **recfrm))
        except BunchPatternNotFound:
            runtype = litfrm.DetectorCtrl.patternType.value
            errmsg = f"BunchPattern not found ({runtype})"
            writer.writerow(dict(runno=runno, exception=errmsg))
            print(f"{runno:4d} - skip: {errmsg}")
        except DetectorNotFound:
            errmsg = "Detector not found"
            writer.writerow(dict(runno=runno, exception=errmsg))
            print(f"{runno:4d} - skip: {errmsg}")

    f.close()



# coding: utf-8

__version__ = "0.0.5"

from .lossy.rounding import errb_round
from .litfrm.agipd_base import (
    AgipdLitFrameFinderBase, FramesAnnotation, DESTINATION,
    AGIPD_MAX_CELL
)
from .litfrm.agipd_offline import (
    AgipdLitFrameFinderOffline, AgipdLitFrameFinderMID,
    AgipdLitFrameFinderSPB, AgipdLitFrameFinderHED,
    LitFrameFinderError, BunchPatternNotFound, DetectorNotFound,
    TriggerNotFound, NoReferenceDelay, make_litframe_finder
)



from .agipd_base import (
    AgipdLitFrameFinderBase, FramesAnnotation, DESTINATION,
    AGIPD_MAX_CELL
)
from .agipd_offline import (
    AgipdLitFrameFinderOffline, AgipdLitFrameFinderMID,
    AgipdLitFrameFinderSPB, AgipdLitFrameFinderHED,
    LitFrameFinderError, BunchPatternNotFound, DetectorNotFound,
    TriggerNotFound, NoReferenceDelay, make_litframe_finder
)



import os
import sys
import re
import setuptools
import numpy as np

from distutils.extension import Extension
from Cython.Build import cythonize


def find_version(filename):
    script_path = os.path.dirname(os.path.realpath(__file__))
    version_file = open(os.path.join(script_path, filename)).read()
    match = re.search(r'^__version__ = "(\d+\.\d+\.\d+)"', version_file, re.M)
    if match is not None:
        return match.group(1)
    raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="EXtra-redu",
    version=find_version("src/extra_redu/__init__.py"),
    author="European XFEL GmbH",
    author_email="da-support@xfel.eu",
    maintainer="Egor Sobolev",
    url="https://git.xfel.eu/dataAnalysis/data-reduction",
    description="Tools for XFEL data quality estimation and data reduction",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="BSD-3-Clause",
    packages=setuptools.find_packages('src'),
    package_dir={"": "src"},
    include_package_data=True,
    package_data={
        'extra_redu.lossy': ['*.h', '*.pyx', '*.pxd'],
    },
    ext_modules=cythonize(
        [
            Extension(
                "extra_redu.lossy.rounding",
                [
                    "src/extra_redu/lossy/rounding.pyx",
                    "src/extra_redu/lossy/rounding_impl.c"
                ],
                include_dirs=[np.get_include()],
            )
        ],
        compiler_directives={'language_level' : "3"},
    ),
    setup_requires=[
        'numpy',
        'cython',
    ],
    install_requires=[
        'numpy',
        'matplotlib',
        'euxfel-bunch-pattern>=0.6',
        'extra-data>=1.9.1',
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-cov',
            'mock',
          ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Utilities",
    ],
    python_requires='>=3.6',
)



import numpy as np

from collections import namedtuple
from euxfel_bunch_pattern import (
    DESTINATION_MASK, PHOTON_LINE_DEFLECTION, DESTINATION_T4D, DESTINATION_T5D
)

XgmIntensity = namedtuple('XgmIntensity', ['mu', 'sig'])
ReferenceDelay = namedtuple('ReferenceDelay', [
    'delay', 'pulse', 'frame', 'repetitionRate', 'position'])


class FramesAnnotation:
    """Frames annotatons (metadata)"""

    def __init__(self, npulse, energy, energy_std, master_ids,
                 xgm_ids, hasdata=None):
        self.npulse = npulse
        self.energy = energy
        self.energy_std = energy_std
        self.master_ids = master_ids
        self.xgm_ids = xgm_ids
        hasxray = npulse > 0
        self.hasdata = hasxray if hasdata is None else hasdata
        self.nlit = hasxray.sum()
        self.ndata = self.hasdata.sum()

    @classmethod
    def non_xray(cls, ncell, hasdata=None):
        return cls(
            np.zeros(ncell, np.uint16),
            np.zeros(ncell, float),
            np.zeros(ncell, float),
            np.zeros(0, np.uint16),
            np.zeros(0, np.uint16),
            np.ones(ncell, bool) if hasdata is None else hasdata
        )

    @classmethod
    def dark(cls, ncell):
        return cls.non_xray(ncell, np.zeros(ncell, bool))


class ParamDecoder:
    """Decodes Bunch Pattern"""

    @classmethod
    def get_instrument_pulse_mask(cls, bunch):
        return (bunch & cls.mask) == cls.code

    @classmethod
    def get_xgm_pulse_mask(cls, bunch):
        xgm_pulses = (bunch & cls.mask_xgm) == cls.code_xgm
        inst_pulses = (bunch & cls.mask) == cls.code
        return inst_pulses[xgm_pulses]

    @classmethod
    def get_xgm_pulse_mask_in_file(cls, bunch, xgm_maxlen=1000):
        xgm_pulses = (bunch & cls.mask_xgm) == cls.code_xgm
        inst_pulses = (bunch & cls.mask) == cls.code
        return inst_pulses[xgm_pulses]


class SA1(ParamDecoder):
    """SASE 1 Bunch Pattern Decoder"""
    mask = PHOTON_LINE_DEFLECTION | DESTINATION_MASK
    code = DESTINATION_T4D
    mask_xgm = DESTINATION_MASK
    code_xgm = DESTINATION_T4D


class SA2(ParamDecoder):
    """SASE 2 Bunch Pattern Decoder"""
    mask = DESTINATION_MASK
    code = DESTINATION_T5D
    mask_xgm = DESTINATION_MASK
    code_xgm = DESTINATION_T5D


class SA3(ParamDecoder):
    """SASE 3 Bunch Pattern Decoder"""
    mask = PHOTON_LINE_DEFLECTION | DESTINATION_MASK
    code = PHOTON_LINE_DEFLECTION | DESTINATION_T4D
    mask_xgm = DESTINATION_MASK
    code_xgm = DESTINATION_T4D


DESTINATION = {
    'SPB': SA1, 'FXE': SA1,
    'MID': SA2, 'HED': SA2,
    'SQS': SA3, 'SCS': SA3,
}

AGIPD_MAX_CELL = 352
CLOCKS_PER_PULSE = 22  # AGIPD timing: 1 clock = 10 ns
RATE_TO_STRIDE = {'4.5': 1, '2.2': 2, '1.1': 4, '0.5': 8}
RATE_OPTIONS = [float(k) for k in RATE_TO_STRIDE]
SPEED_OF_LIGHT = 2.76708438734  # m/tick (1 tick = 9.23 ns)
TICS_PER_PULSE = 24  # TIMESERVER timing: 1 tic = 9.23 ns
EVENT_CODE = {"dynamic": 181, "static": 16}
EVENT_PULSE = {
    16: lambda pulses: 0,
    181: lambda pulses: np.argmax(pulses),
}


class Agipd1g:
    def __init__(self, detector_control_id, litfrm):
        self.detector_control_id = detector_control_id
        self.litfrm = litfrm

    @property
    def state(self):
        return self.det.masterState

    @property
    def exp_type(self):
        return self.det.patternType

    def get_bunch_structure(self):
        """Reads AGIPD frame settings."""
        ncell = self.det.bunchStructure.nPulses.value
        rate = self.det.bunchStructure.repetitionRate.value
        first_pulse_id = self.det.bunchStructure.firstPulse.value
        if ncell > AGIPD_MAX_CELL:
            ncell = AGIPD_MAX_CELL

        return ncell, first_pulse_id, RATE_TO_STRIDE[f"{rate:.1f}"]


class Agipd2g:
    def __init__(self, detector_control_id, litfrm):
        self.detector_control_id = detector_control_id
        self.litfrm = litfrm
        if litfrm.alignMethod.value == 'by reference delay':
            raise ValueError(
                "Alignment method 'by reference delay' "
                "is not implemented for AGIPD 2nd gen.")

    @property
    def state(self):
        return self.det.ctrlDevsState

    @property
    def exp_type(self):
        return self.det.expType

    def get_bunch_structure(self):
        """Reads AGIPD frame settings."""
        ncell = self.det.bunchStructure.nPulses.value
        rate = self.det.bunchStructure.repetitionRate.value
        if ncell > AGIPD_MAX_CELL:
            ncell = AGIPD_MAX_CELL

        return ncell, 0, RATE_TO_STRIDE[f"{rate:.1f}"]


AGIPDGEN = {
    "AgipdComposite": Agipd1g,
    "Agipd2Composite": Agipd2g,
}


class AgipdLitFrameFinderBase:
    """Base class containing core LitFrameFinder algorithm"""

    def make_dark_run(self):
        """Make LitFrameFinder response for dark or pc mode."""
        ncell, first_pulse_id, stride = self._agipd.get_bunch_structure()
        agipd_pulse_ids = np.arange(ncell) * stride + first_pulse_id

        frames = FramesAnnotation.non_xray(ncell)
        return 0, ncell, agipd_pulse_ids, frames

    def process_train(self, bp, xgm):
        """Process train data and make LitFrameFinder response."""

        ncell, first_pulse_id, stride = self._agipd.get_bunch_structure()
        agipd_pulse_ids = np.arange(ncell) * stride + first_pulse_id

        bunch_pattern = bp.data.data.bunchPatternTable.value
        # ts = meta.timestamp.timestamp
        # tid = meta.timestamp.tid

        pulses = self.dst.get_instrument_pulse_mask(bunch_pattern)
        npulse = pulses.sum()

        xgm_pulses = self.dst.get_xgm_pulse_mask(bunch_pattern)
        nxgm = xgm_pulses.size
        if xgm is None:
            intensity = XgmIntensity(
                np.full(nxgm, self.missedIntensityTD, np.float32),
                np.full(nxgm, self.missedIntensitySigmaTD, np.float32),
            )
        else:
            intensity = XgmIntensity(
                xgm.data.data.intensityTD.value[:nxgm][xgm_pulses],
                xgm.data.data.intensitySigmaTD.value[:nxgm][xgm_pulses]
            )

        xgm_indices = np.flatnonzero(xgm_pulses)
        frames = self.get_agipd_frames_annotation(
            npulse, pulses, intensity, xgm_indices, ncell, stride
        )
        # or self.DetectorCtrl.patternTypeIndex != 0
        if self._agipd.exp_type.value != 'XRay':
            frames.hasdata[:] = True
            frames.ndata = frames.hasdata.size

        return npulse, ncell, agipd_pulse_ids, frames

    def get_pulse_to_frame_alignment(self, pulses, stride):
        """Returns the reference pulse and the shift in pulses
           to the first exposed pulse slot.
        """
        if self.alignMethod.value == 'by first pulse':
            shift = self.referenceFrame.value * stride
            ref_pulse = np.argmax(pulses)
        elif self.alignMethod.value == 'by reference delay':
            frame_timestep = RATE_TO_STRIDE[
                f"{self.ref.repetitionRate:.1f}"] * TICS_PER_PULSE
            shift = (self.ref.delay + self.ref.frame * frame_timestep
                     - self.trigger_delay)
            if self.motor_position is not None:
                unit_coef = self.motor.unitConversionCoef.value
                shift -= (unit_coef * (self.ref.position
                                       - self.motor_position) / SPEED_OF_LIGHT)

            shift = round(shift / TICS_PER_PULSE)
            ref_pulse = EVENT_PULSE[
                self.trigger_event](pulses) + self.ref.pulse
        else:
            raise ValueError("Unexpected value of attribute "
                             f"alignMethod: {self.alignMethod.value}")

        return ref_pulse, shift

    def get_agipd_frames_annotation(
        self, npulse, pulses, intensity, xgm_indices, ncell, stride
    ):
        """Matches X-ray pulses, their energy, etc to the AGIPD frames."""
        # if no beam
        if npulse == 0 or not self.shutters_are_open:
            return FramesAnnotation.dark(ncell)

        # read exposure time
        exposure_in_tics = self.DetectorCtrl.integrationTime.value
        exposure = min(int(np.ceil(exposure_in_tics
                                   / CLOCKS_PER_PULSE)), stride)

        pulse_no = np.full(pulses.size, 65535, dtype=np.uint16)
        pulse_no[pulses] = np.arange(npulse, dtype=np.uint16)

        # compute pulse per frame pattern
        nfrm_in_pulses = ncell * stride
        pulse_to_frame = np.zeros(nfrm_in_pulses, dtype=bool)
        energy_to_frame = np.zeros(nfrm_in_pulses, dtype=float)
        energy_std_to_frame = np.zeros(nfrm_in_pulses, dtype=float)
        local_pulse_ids = np.full(nfrm_in_pulses, 65535, dtype=np.uint16)

        ref_pulse, shift = self.get_pulse_to_frame_alignment(pulses, stride)

        # compute slices of frame and pulse arrays
        first_pulse = np.argmax(pulses)
        p0 = max(first_pulse - shift, 0)
        f0 = (shift - first_pulse) * (shift > first_pulse)

        npulse_exposed = min(2700 - p0, nfrm_in_pulses - f0)
        pN = p0 + npulse_exposed
        fN = f0 + npulse_exposed

        pulse_to_frame[f0:fN] = pulses[p0:pN]
        pulse_no_to_frame = np.extract(pulses[p0:pN], pulse_no[p0:pN])

        local_pulse_ids[pulse_to_frame] = pulse_no_to_frame
        local_pulse_ids = local_pulse_ids.reshape(ncell, stride)[:, :exposure]
        local_pulse_ids = local_pulse_ids[local_pulse_ids != 65535]

        master_pulse_ids = np.flatnonzero(pulses)[local_pulse_ids]
        xgm_pulse_ids = xgm_indices[local_pulse_ids]

        energy_to_frame[pulse_to_frame] = intensity.mu[pulse_no_to_frame]
        energy_to_frame = energy_to_frame.reshape(ncell, stride)[:, :exposure]

        energy_std_to_frame[pulse_to_frame] = intensity.sig[pulse_no_to_frame]
        energy_std_to_frame = energy_std_to_frame.reshape(
            ncell, stride)[:, :exposure]

        pulse_to_frame = pulse_to_frame.reshape(ncell, stride)[:, :exposure]

        frames = FramesAnnotation(
            pulse_to_frame.astype(np.uint16).sum(1),
            energy_to_frame.sum(1),
            np.sqrt(np.square(energy_std_to_frame).sum(1)),
            master_pulse_ids,
            xgm_pulse_ids,
        )
        return frames



import numpy as np
import matplotlib.pyplot as plt


def draw_cells(cflag):
    """Plots lit/dark cells"""
    Y, X = np.mgrid[0:12, 0:33]
    C = np.pad(
        cflag.astype(int), [0, 352-cflag.size],
        constant_values=7
    ).reshape(11, 32)

    fig, ax = plt.subplots(
        1, 1, figsize=(10, 4), tight_layout=True, dpi=200
    )
    ax.pcolor(
        X-0.5, Y-0.5, C, cmap=plt.cm.tab10,
        edgecolors='w', lw=0.5, vmin=0, vmax=9
    )
    ax.set_aspect(1)
    ax.set_title('Lit cells', fontsize=8)
    ax.tick_params(axis='both', labelsize=8)
    ax.yaxis.set_ticks(range(0, 11))
    ax.yaxis.set_ticklabels(range(0, 321, 32))
    ax.xaxis.set_ticks(range(3, 33, 4))

    plt.show()
